@@ -1,7 +1,7 @@
 from datetime import datetime
 import json
 import logging
-from typing import Any, cast, Dict, Iterable, List, Optional, Union
+from typing import Any, cast, Dict, List, Optional, Union
 
 import aporia
 from aporia.consts import LOGGER_NAME
@@ -25,6 +25,7 @@ class BaseModel:
         event_loop: Optional[EventLoop],
         debug: bool,
         throw_errors: bool,
+        timeout: Optional[int],
     ):
         """Initializes a BaseModel object.
 
@@ -36,6 +37,7 @@ class BaseModel:
             event_loop (Optional[EventLoop]): AsyncIO event loop
             debug (bool): True to enable debug mode
             throw_errors (bool): True to raise exceptions in object creation and set_features
+            timeout (Optional[int]): Timeout, in seconds, for synchronous functions
         """
         logger.debug(f"Initializing model object for model {model_id} version {model_version}")
         self.model_id = model_id
@@ -47,6 +49,7 @@ class BaseModel:
         self._throw_errors = throw_errors
         self._model_ready = False
         self._model_version_exists = False
+        self._timeout = cast(int, timeout)
 
         # Check model existence
         if event_loop is not None and graphql_client is not None:
@@ -78,7 +81,8 @@ class BaseModel:
         }
 
         try:
-            result = (await self._graphql_client.query(query, variables))["modelExists"]
+            result = await self._graphql_client.query(query, variables, self._timeout)
+            result = result["modelExists"]
             if not result["modelExists"]:
                 raise AporiaError(f"Model {self.model_id} does not exist.")
 
@@ -94,42 +98,30 @@ class BaseModel:
             )
 
     def set_features(
-        self, feature_names: List[str], categorical: Optional[Iterable[Union[str, int]]] = None,
+        self, feature_names: List[str], categorical: Optional[List[Union[int, str]]] = None,
     ):
         """Sets feature names and categorical features for the model.
 
         Args:
             feature_names (List[str]): List of feature names
-            categorical (Optional[Iterable[Union[str, int]]], optional): List of categorical
-                features - either names or indexes. Defaults to None.
+            categorical (List[Union[int, str]], optional): A list indicating which of the
+                features in the 'feature_names' argument are considered categorical.
         """
         logger.debug("Setting model features.")
         if not self._model_ready:
             return
 
-        categorical_features = []
-        if categorical is not None:
-            for feature in categorical:
-                if isinstance(feature, str):
-                    categorical_features.append(feature_names.index(feature))
-                else:
-                    categorical_features.append(feature)
+        self._event_loop.run_coroutine(self._set_features(feature_names, categorical))
 
-        self._event_loop.run_coroutine(self._set_features(feature_names, categorical_features))
-
-    async def _set_features(self, feature_names: List[str], categorical_features: List[int]):
-        """Internal, asynchronous implementation for set_features().
-
-        Args:
-            feature_names (List[str]): List of feature names
-            categorical_features (List[int]): Categorical feature indexes
-        """
+    async def _set_features(
+        self, feature_names: List[str], categorical_features: Optional[List[Union[int, str]]]
+    ):
         query = """
             mutation SetFeatures(
                 $modelId: String!,
                 $modelVersion: String!,
                 $featureNames: [String]!,
-                $categoricalFeatures: [Int]!
+                $categoricalFeatures: [FeatureDescriptor]!
             ) {
                 setFeatures(
                     modelId: $modelId,
@@ -137,7 +129,7 @@ class BaseModel:
                     featureNames: $featureNames,
                     categoricalFeatures: $categoricalFeatures
                 ) {
-                    ok
+                    warnings
                 }
             }
         """
@@ -150,7 +142,10 @@ class BaseModel:
         }
 
         try:
-            await self._graphql_client.query(query, variables)
+            result = await self._graphql_client.query(query, variables, self._timeout)
+            for warning in result["setFeatures"]["warnings"]:
+                logger.warning(warning)
+
             # If the query didn't fail, then the model version should now exist in the DB
             self._model_version_exists = True
         except Exception as err:
@@ -161,71 +156,97 @@ class BaseModel:
                 original_exception=err,
             )
 
-    def add_input_metadata(self, metadata: Dict[str, str]):
-        """Adds input metadata fields to the model.
+    def add_extra_inputs(self, extra_inputs: Dict[str, str]):
+        """Adds extra inputs to the model, which can then be reported in log_predict.
 
         Args:
-            metadata (Dict[str, str]): A dict of field_name: field_type metadata fields. field
-                type must be either 'numeric' or 'categorical'.
+            extra_inputs (Dict[str, str]): A mapping of input_name: input_type. type must
+                be one of the following: 'numeric', 'categorical', 'boolean', 'string'.
         """
-        self._event_loop.run_coroutine(self._add_metadata(metadata, "INPUT"))
+        self._event_loop.run_coroutine(self._add_extra_inputs(extra_inputs))
 
-    def add_output_metadata(self, metadata: Dict[str, str]):
-        """Adds output metadata fields to the model.
-
-        Args:
-            metadata (Dict[str, str]): A dict of field_name: field_type metadata fields. field
-                type must be either 'numeric' or 'categorical'.
-        """
-        self._event_loop.run_coroutine(self._add_metadata(metadata, "OUTPUT"))
-
-    async def _add_metadata(self, metadata: Dict[str, str], designation: str):
+    async def _add_extra_inputs(self, extra_inputs: Dict[str, str]):
         try:
             query = """
-                mutation AddMetadata(
+                mutation AddExtraInputs(
                     $modelId: String!,
                     $modelVersion: String!,
-                    $metadata: [MetadataFieldInput]!,
+                    $extraInputs: [ExtraInput]!,
                 ) {
-                    addMetadata(
+                    addExtraInputs(
                         modelId: $modelId,
                         modelVersion: $modelVersion,
-                        metadata: $metadata,
+                        extraInputs: $extraInputs,
                     ) {
                         warnings
-                        overwrittenFields {
-                            name
-                            type
-                        }
                     }
                 }
             """
 
-            metadata_fields: List[Dict[str, str]] = []
-            for value_name, value_type in metadata.items():
-                metadata_fields.append(
-                    {"name": value_name, "type": value_type.upper(), "designation": designation}
-                )
-
             variables = {
                 "modelId": self.model_id,
                 "modelVersion": self.model_version,
-                "metadata": metadata_fields,
+                "extraInputs": [
+                    {"name": input_name, "type": input_type.upper()}
+                    for input_name, input_type in extra_inputs.items()
+                ],
             }
 
-            result = (await self._graphql_client.query(query, variables))["addMetadata"]
-            for field in result["overwrittenFields"]:
-                logger.warning(
-                    f"The type of metadata field {field['name']} was overwritten "
-                    f"from {metadata[field['name']]} to {field['type']}."
-                )
-
-            for warning in result["warnings"]:
+            result = await self._graphql_client.query(query, variables, self._timeout)
+            for warning in result["addExtraInputs"]["warnings"]:
                 logger.warning(warning)
 
         except Exception as err:
             handle_error(
-                message=f"Adding {designation} metadata failed, error: {str(err)}",
+                message=f"Adding extra inputs failed, error: {str(err)}",
+                add_trace=self._debug,
+                raise_exception=self._throw_errors,
+                original_exception=err,
+            )
+
+    def add_extra_outputs(self, extra_outputs: Dict[str, str]):
+        """Adds extra outputs to the model, which can then be reported in log_predict.
+
+        Args:
+            extra_outputs (Dict[str, str]): A mapping of output_name: output_type. type
+                must be one of the following: 'numeric', 'categorical', 'boolean', 'string'.
+        """
+        self._event_loop.run_coroutine(self._add_extra_outputs(extra_outputs))
+
+    async def _add_extra_outputs(self, extra_outputs: Dict[str, str]):
+        try:
+            query = """
+                mutation AddExtraOutputs(
+                    $modelId: String!,
+                    $modelVersion: String!,
+                    $extraOutputs: [ExtraOutput]!,
+                ) {
+                    addExtraOutputs(
+                        modelId: $modelId,
+                        modelVersion: $modelVersion,
+                        extraOutputs: $extraOutputs,
+                    ) {
+                        warnings
+                    }
+                }
+            """
+
+            variables = {
+                "modelId": self.model_id,
+                "modelVersion": self.model_version,
+                "extraOutputs": [
+                    {"name": output_name, "type": output_type.upper()}
+                    for output_name, output_type in extra_outputs.items()
+                ],
+            }
+
+            result = await self._graphql_client.query(query, variables, self._timeout)
+            for warning in result["addExtraOutputs"]["warnings"]:
+                logger.warning(warning)
+
+        except Exception as err:
+            handle_error(
+                message=f"Adding extra outputs failed, error: {str(err)}",
                 add_trace=self._debug,
                 raise_exception=self._throw_errors,
                 original_exception=err,
@@ -237,7 +258,8 @@ class BaseModel:
         y: List[float],
         confidence: Optional[List[float]] = None,
         occurred_at: Optional[datetime] = None,
-        **kwargs,
+        extra_inputs: Optional[Dict[str, FeatureValue]] = None,
+        extra_outputs: Optional[Dict[str, FeatureValue]] = None,
     ):
         """Logs a single prediction.
 
@@ -246,23 +268,28 @@ class BaseModel:
             y (List[float]): Prediction result
             confidence (List[float], optional): Prediction confidence. Defaults to None.
             occurred_at (datetime, optional): Prediction timestamp. Defaults to None.
-            kwargs (Dict[str, FeatureValue]): Metadata fields, each field must be a single bool,
-                int, float or string
+            extra_inputs (Dict[str, FeatureValue], optional): Extra inputs, each input
+                value must be a single bool, int, float or string. Defaults to None.
+            extra_outputs (Dict[str, FeatureValue], optional): Extra outputs, each output
+                value must be a single bool, int, float or string. Defaults to None.
 
         Notes:
+            * The features reported in the x parameter must be set with the set_features
+              method before logging predictions.
+            * Extra inputs and outputs should be defined with add_extra_inputs or
+              add_extra_outputsbefore they are passed to log_predict.
             * Many libraries provide a way to convert their objects to lists:
                 * numpy array: Use data.tolist()
                 * pandas dataframe: Use data.values.tolist()
                 * scipy sparse matrix: Use data.toarray().tolist()
-            * Metadata fields should be defined with add_input_metadata or add_output_metadata
-              before they are passed to log_predict
         """
         self.log_predict_batch(
             x=[x],
             y=[y],
             confidence=None if confidence is None else [confidence],
             occurred_at=occurred_at,
-            **{name: [value] for name, value in kwargs.items()},
+            extra_inputs=None if extra_inputs is None else [extra_inputs],
+            extra_outputs=None if extra_outputs is None else [extra_outputs],
         )
 
     def log_predict_batch(
@@ -271,7 +298,8 @@ class BaseModel:
         y: List[List[float]],
         confidence: Optional[List[List[float]]] = None,
         occurred_at: Optional[datetime] = None,
-        **kwargs,
+        extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
+        extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
     ):
         """Logs multiple predictions.
 
@@ -280,16 +308,22 @@ class BaseModel:
             y (List[List[float]]): Predictions results for each prediction
             confidence (List[List[float]], optional): Confidence for each prediction. Defaults to None.
             occurred_at (datetime, optional): Timestamp for each prediction. Defaults to datetime.now().
-            kwargs (Dict[str, List[FeatureValue]]): Metadata fields, each field must be a list of bool,
-                int, float or string values, one value for each prediction in the batch
+            extra_inputs (List[Dict[str, FeatureValue]], optional): Extra inputs for each prediction, each
+                element in the list should a dict mapping the input name to its value, which
+                can be bool, int, float or string. Defaults to None.
+            extra_outputs (List[Dict[str, FeatureValue]], optional): Extra outputs for each prediction, each
+                element in the list should a dict mapping the output name to its value, which
+                can be bool, int, float or string. Defaults to None.
 
         Notes:
+            * The features reported in the x parameter must be set with the set_features
+              method before logging predictions.
+            * Extra inputs and outputs should be defined with add_extra_inputs or
+              add_extra_outputsbefore they are passed to log_predict_batch.
             * Many libraries provide a way to convert their objects to lists:
                 * numpy array: Use data.tolist()
                 * pandas dataframe: Use data.values.tolist()
                 * scipy sparse matrix: Use data.toarray().tolist()
-            * Metadata fields should be defined with add_input_metadata or add_output_metadata
-              before they are passed to log_predict_batch
         """
         logger.debug(f"Logging {len(y)} predictions")
         if not self._model_ready:
@@ -301,18 +335,21 @@ class BaseModel:
             )
             return
 
-        metadata = None
-        if len(kwargs) > 0:
-            metadata = kwargs
-
-        if not self._is_valid_predict_input(x=x, y=y, confidence=confidence, metadata=metadata):
-            logger.info("Logging prediction failed: Invalid input.")
+        if not self._is_valid_predict_input(
+            x=x, y=y, confidence=confidence, extra_inputs=extra_inputs, extra_outputs=extra_outputs
+        ):
+            logger.warning("Logging prediction failed: Invalid input.")
             return
 
         try:
             self._event_loop.run_coroutine(
                 coro=self._log_predict_batch(
-                    x=x, y=y, confidence=confidence, occurred_at=occurred_at, metadata=metadata
+                    x=x,
+                    y=y,
+                    confidence=confidence,
+                    occurred_at=occurred_at,
+                    extra_inputs=extra_inputs,
+                    extra_outputs=extra_outputs,
                 ),
                 await_result=False,
             )
@@ -331,17 +368,9 @@ class BaseModel:
         y: List[List[float]],
         confidence: Optional[List[List[float]]] = None,
         occurred_at: Optional[datetime] = None,
-        metadata: Optional[Dict[str, List[FeatureValue]]] = None,
+        extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
+        extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
     ):
-        """Internal, asynchronous implementation ofr log_predict_batch.
-
-        Args:
-            x (List[List[FeatureValue]]): X values for all of the features in each of the predictions
-            y (List[List[float]]): Predictions results for each prediction
-            confidence (List[List[float]], optional): Confidence for each prediction. Defaults to None.
-            occurred_at (datetime, optional): Timestamp for each prediction. Defaults to datetime.now().
-            metadata (Dict[str, List[FeatureValue]], optional): Metadata values. Defaults to None.
-        """
         try:
             query = """
                 mutation LogPredict(
@@ -352,7 +381,8 @@ class BaseModel:
                     $confidence: [[Float]],
                     $occurredAt: String,
                     $environment: String!,
-                    $metadata: [MetadataValues]
+                    $extraInputs: [[ExtraInputValue]]
+                    $extraOutputs: [[ExtraOutputValue]]
                 ) {
                     logPredictions(
                         modelId: $modelId,
@@ -362,19 +392,14 @@ class BaseModel:
                         confidence: $confidence,
                         occurredAt: $occurredAt,
                         environment: $environment,
-                        metadata: $metadata
+                        extraInputs: $extraInputs
+                        extraOutputs: $extraOutputs
                     ) {
                         warnings
                     }
                 }
             """
 
-            metadata_values = None
-            if metadata is not None:
-                metadata_values = [
-                    {"name": name, "values": values} for name, values in metadata.items()
-                ]
-
             variables = {
                 "modelId": self.model_id,
                 "modelVersion": self.model_version,
@@ -383,11 +408,12 @@ class BaseModel:
                 "confidence": confidence,
                 "occurredAt": None if occurred_at is None else occurred_at.isoformat(),
                 "environment": self._environment,
-                "metadata": metadata_values,
+                "extraInputs": self._build_extra_io_values(extra_inputs),
+                "extraOutputs": self._build_extra_io_values(extra_outputs),
             }
 
-            result = (await self._graphql_client.query(query, variables))["logPredictions"]
-            for warning in result["warnings"]:
+            result = await self._graphql_client.query_with_retries(query, variables)
+            for warning in result["logPredictions"]["warnings"]:
                 logger.warning(warning)
 
         except Exception as err:
@@ -399,12 +425,26 @@ class BaseModel:
                 log_level=logging.INFO if isinstance(err, AporiaError) else logging.ERROR,
             )
 
+    @staticmethod
+    def _build_extra_io_values(
+        data: Optional[List[Dict[str, FeatureValue]]] = None
+    ) -> Optional[List[List[dict]]]:
+        if data is None:
+            return None
+
+        result = []
+        for data_point in data:
+            result.append([{"name": name, "value": value} for name, value in data_point.items()])
+
+        return result
+
     @staticmethod
     def _is_valid_predict_input(
         x: List[List[FeatureValue]],
         y: List[List[float]],
         confidence: Optional[List[List[float]]] = None,
-        metadata: Optional[Dict[str, List[FeatureValue]]] = None,
+        extra_inputs: Optional[List[Dict[str, FeatureValue]]] = None,
+        extra_outputs: Optional[List[Dict[str, FeatureValue]]] = None,
     ) -> bool:
         if not Model._is_valid_predict_param_list(x):
             logger.debug("Invalid input format for x parameter")
@@ -427,24 +467,36 @@ class BaseModel:
                 logger.debug("Invalid input: y and confidence should have identical length")
                 return False
 
-        if metadata is not None:
-            if not Model._is_valid_metadata(metadata=metadata, expected_length=len(x)):
+        if extra_inputs is not None:
+            if not Model._is_valid_extra_io(data=extra_inputs, expected_length=len(x)):
                 logger.debug(
-                    "Invalid input: each metadata value must be a list with values "
-                    "for each prediction in the batch"
+                    "Invalid input: extra_inputs must be a list of values equal in length to the "
+                    "number of predictions, such that each element in the list is a dict "
+                    "containing the extra inputs for a single prediction"
+                )
+                return False
+
+        if extra_outputs is not None:
+            if not Model._is_valid_extra_io(data=extra_outputs, expected_length=len(x)):
+                logger.debug(
+                    "Invalid input: extra_outputs must be a list of values equal in length to the "
+                    "number of predictions, such that each element in the list is a dict "
+                    "containing the extra outputs for a single prediction"
                 )
                 return False
 
         return True
 
     @staticmethod
-    def _is_valid_metadata(metadata: Dict[str, List[FeatureValue]], expected_length: int) -> bool:
-        for metadata_values in metadata.values():
-            if not isinstance(metadata_values, list):
-                return False
+    def _is_valid_extra_io(data: List[Dict[str, FeatureValue]], expected_length: int) -> bool:
+        if not isinstance(data, list):
+            return False
 
-            elif len(metadata_values) != expected_length:
-                return False
+        if len(data) != expected_length:
+            return False
+
+        if not all(isinstance(data_point, dict) for data_point in data):
+            return False
 
         return True
 
@@ -506,7 +558,7 @@ class BaseModel:
                         environment: $environment,
                         data: $data
                     ) {
-                        ok
+                        warnings
                     }
                 }
             """
@@ -518,7 +570,10 @@ class BaseModel:
                 "data": json.dumps(data),
             }
 
-            await self._graphql_client.query(query, variables)
+            result = await self._graphql_client.query_with_retries(query, variables)
+            for warning in result["logArbitraryData"]["warnings"]:
+                logger.warning(warning)
+
         except Exception as err:
             handle_error(
                 message=f"Logging arbitrary data failed, error: {str(err)}",
@@ -536,8 +591,9 @@ class Model(BaseModel):
         """Initializes a Model object.
 
         Args:
-            model_id (str): Model identifier.
-            model_version (str): Model version
+            model_id (str): Model identifier, as received from the Aporia dashboard.
+            model_version (str): Model version - this can be any string that represents the model
+                version, such as "v1" or a git commit hash.
         """
         if aporia.context is None:
             logger.error("Aporia was not initialized.")
@@ -549,6 +605,7 @@ class Model(BaseModel):
                 event_loop=None,
                 debug=False,
                 throw_errors=False,
+                timeout=None,
             )
         else:
             super().__init__(
@@ -559,4 +616,5 @@ class Model(BaseModel):
                 environment=aporia.context.environment,
                 debug=aporia.context.debug,
                 throw_errors=aporia.context.throw_errors,
+                timeout=aporia.context.timeout,
             )
